#!/usr/bin/env python
"""
Hiveary
https://hiveary.com

Licensed under Simplified BSD License (see LICENSE)
(C) Hiveary, LLC 2013 all rights reserved
"""

import datetime
import json
import kombu
import kombu.common
import logging
import oauth2
import os
import platform
import random
import socket
from ssl import CERT_REQUIRED, CERT_NONE
import subprocess
import sys
import time
import traceback
import urllib2

# Windows specific imports
if subprocess.mswindows:
  import pythoncom
  import wincom

# Local imports
from . import oauth_client
from . import paths
from . import sysinfo


class NetworkController(object):
  """Class used to handle network functions, such as the AMQP connection."""

  PING_TIMER = 120  # How often to ping the server, in seconds
  MAX_BACKOFF_MULTIPLE = 10

  def __init__(self, reactor=None, logger=None):
    """Initialze the controller.

    Args:
      reactor: A reference to the twisted reactor controlling the agent.
      logger: A logging object to use.
    """

    self.logger = logger or logging.getLogger('hiveary_agent.network')

    # Configuration values - most of these will be modified by the host's
    # configuration file
    self.running = True
    self.owner = ''
    self.hostname = ''
    self.access_token = 'hunter5'
    self.debug_mode = False
    self.disable_ssl_verification = False
    self.obj_id = None
    self.remote_host = ''
    self.amqp_server = ''
    self.ca_bundle = None
    self.current_system = platform.system()

    # AMQP connection values
    self.user_id = None
    self.amqp_password = None
    self.amqp = None

    # Thread and deferred management
    self.reactor = reactor

    self.expected_values = {}

  def ensure_internet_connection(self, test_url='http://198.41.189.27'):
    """Blocks until there is an active connection to the public internet.
    Uses an IP address by default, since DNS lookups can cause urllib2 to
    block for a while and eat up system resources while doing so."""

    self.logger.debug('Checking for an active internet connection...')

    while True:
      try:
        urllib2.urlopen(test_url, timeout=1)
        break
      except urllib2.URLError:
        self.logger.debug('Test connection failed:\n', exc_info=True)
        time.sleep(5)

    self.logger.debug('Active connection found!')

  def initialize_amqp(self):
    """Method to establish an AMQP connection and consumers."""

    # Find out the credentials used for AMQP
    url = 'https://{host}/amqp/account?hostname={hostname}'.format(
        host=self.remote_host, hostname=self.hostname)
    resp, content = self.request_with_backoff(url, method='GET')

    if resp.status == 200:
      self.logger.info('Retrieved AMQP credentials')
      content_json = json.loads(content)
      self.amqp_password = content_json.get('amqp_password')
      self.user_id = content_json.get('user_id')
      self.obj_id = content_json.get('host_id')
    elif resp.status == 409:
      self.logger.error('Unable to connect this host to the server, '
                        'you\'ve already used all of your licenses. '
                        'Please vist https://%s/upgrade to upgrade your plan.',
                        self.remote_host)
      sys.exit(409)
    else:
      self.logger.error('Failed to retrieve AMQP credentials. Status: %s',
                        resp.status)
      sys.exit(resp.status)

    if self.obj_id is None or self.amqp_password is None or self.user_id is None:
      self.logger.error('Missing required parameters to establish an AMQP connection')
      sys.exit(1)

    # Create the connection
    ca_certs = os.path.join(paths.get_program_path(), 'ca-bundle.pem')
    self.logger.debug('Using SSL cert bundle at "%s"', ca_certs)

    ssl_options = {
        'ca_certs': ca_certs,
        'cert_reqs': CERT_REQUIRED,
      }
    # Disable certificate validation when debugging a non-frozen app or when
    # specifically told to as a startup arg
    if self.disable_ssl_verification:
      ssl_options['cert_reqs'] = CERT_NONE

    self.logger.debug('Connecting to %s as user %s', self.amqp_server, self.user_id)
    self.amqp = kombu.Connection(self.amqp_server, self.user_id, self.amqp_password,
                                 port=5671, ssl=ssl_options, insist=True)

    self.amqp.ensure_connection(errback=self.amqp_errback, interval_max=60)
    self.logger.info('SSL-AMQP connection established')
    self.reactor.callInThread(self.drain_events)

  def stop_amqp(self):
    """Stops listening for AMQP messages and releases the connection."""

    self.running = False
    if self.amqp:
      try:
        self.amqp.release()
      except:
        pass

  def amqp_reconnect(self):
    """Handles re-establishing an AMQP connection after an error occurred.
    In some error cases, the connection will be marked as connected despite
    the error, so ensure_connection will not be enoguh to reconnnect.
    Forcing a release first resolves this."""

    self.amqp.release()
    self.amqp.ensure_connection(errback=self.amqp_errback,
                                interval_start=5,
                                interval_step=5,
                                interval_max=60)

  def drain_events(self):
    """Attempts to receive a message from the AMQP server and times out after
    a short wait. Loops until the agent is marked as stopping. This allows us
    to gracefully close the connection when waiting for messages."""

    # Setup the consumers
    task_queue = kombu.Queue('agent.{user}.tasks.{host}'.format(
                             user=self.user_id, host=self.obj_id))
    try:
      task_consumer = self.amqp.Consumer(task_queue, auto_declare=False,
                                         callbacks=[self.task_callback])
      task_consumer.consume()
    except:
      if self.running:
        raise

    loop = kombu.common.eventloop(self.amqp, timeout=1, ignore_timeouts=True)
    self.logger.info('Draining events from the server')

    while self.running:
      try:
        next(loop)
      except Exception, err:
        # Errors generated while the agent is stopping can be ignored
        if self.running:
          self.logger.error('AMQP error while draining events: %s', err)
          self.amqp_reconnect()
          break
    else:
      return

    # If the loop broke while the agent is still running, try starting over
    self.drain_events()

  def create_oauth_client(self):
    """Generates an OAuth client that handles the OAuth signature and header.

    Returns:
      The configured OAuth client, an instance of oauth2.Consumer.
    """

    consumer = oauth2.Consumer(key=self.owner, secret=self.access_token)
    try:
      client = oauth_client.OAuthClient(consumer, debug=self.debug_mode,
                                        disable_ssl_verification=self.disable_ssl_verification,
                                        ca_bundle=self.ca_bundle)
    except IOError:
      sys.exit(1)

    return client

  def request_with_backoff(self, url, attempt=0, **kwargs):
    """Makes an OAuth authenticated HTTPS request. If the request fails, it
    will be retried using an exponential backoff.

    Args:
      url: The absolute url of the server.
      attempt: The current attempt. Anything higher than 0 indicates that a
               previous attempt failed and the request is being retried. This
               is used as a multiple to determine how long to backoff.
      kwargs: Any extra arguments to pass to the client.
    Returns:
      A tuple of (response, content), the first being an instance of the
      'httplib2.Http.Response' class, the second being a string that
      contains the response entity body.
    """

    client = self.create_oauth_client()

    try:
      response = client.request(url, **kwargs)

      # Response is a 2-item tuple of headers, content
      self.logger.debug('Got back a response of %s', response[0].status)

      return response
    except socket.error:
      # Parse out extremely verbose/sensitive data
      logged_kwargs = kwargs.copy()
      if 'body' in logged_kwargs:
        del(logged_kwargs.body)

      self.logger.error('Socket error when attempting to send to %s with params %s:',
                        url, logged_kwargs, exc_info=traceback.format_exc())
      self.logger.debug('Verbose parameter information for errored request:\n%s',
                        kwargs)

      timer = (2 ** attempt) + (random.randint(0, 1000) / 1000.0)
      self.logger.error('Retrying in %.3f', timer)

      # Try again after the backoff and increment the attempts made
      time.sleep(timer)

      # Limit the upper end of the backoff timer
      if attempt < self.MAX_BACKOFF_MULTIPLE:
        attempt += 1

      self.ensure_internet_connection()

      return self.request_with_backoff(url, attempt, **kwargs)

  def publish_alert_message(self, alert=None):
    """Publishes an AMQP message to the alert queue, with some extra data.

    Args:
      alert: The alert to publish. This should be an instance of alerts.BaseAlert
          and will be JSONified before being sent.
    """

    alert['host_id'] = self.obj_id
    self.publish_info_message('alert', json.dumps(alert))

  def publish_info_message(self, destination, message='', retry=True):
    """Method to publish an AMQP message.

    Args:
      destination: The AMQP routing key.
      message: The message to publish, as a string (usually a JSONified object).
    """

    self.logger.debug('Sending "%s" AMQP message', destination)

    exchange = kombu.Exchange('agent.{user}'.format(user=self.user_id))
    with self.amqp.Producer(exchange=exchange, routing_key=destination,
                            auto_declare=False) as producer:
      try:
        producer.publish(message, user_id=self.user_id,
                         timestamp=datetime.datetime.utcnow())
      except Exception, err:
        self.logger.error('Error while publishing to AMQP: %s', err)
        self.amqp_reconnect()

        # Retry publishing the message if requested
        if retry:
          self.publish_info_message(destination, message)

  def amqp_errback(self, exc, interval):
    """Error callback fired when there is a problem with the connection or channel.

    Args:
      exc: The exception that occurred.
      interval: Amount of time before the message is attempted again.
    """

    self.logger.error('AMQP connection error: %r. Retrying in %ds', exc, interval)
    self.ensure_internet_connection()

  def ping_pong(self):
    """Function to alert the server that we're still alive and doing science."""

    # Send the ping
    data = {'host_id': self.obj_id}
    self.logger.debug('Sending ping to server with a body of: %s', data)
    self.publish_info_message('ping', json.dumps(data), retry=False)

  def task_callback(self, body, message):
    """Callback for when a message is received from the tasks queue.

    Args:
      body: A string of the response sent back from the broker.
      message: The AMQP object containing metadata about the response.
    """

    self.logger.debug('Received the following task message: %s', body)

    try:
      data = json.loads(body)
    except ValueError:
      self.logger.error('Unable to process task:', exc_info=True)
    else:
      self.run_task(data)

    # Let the broker know that we're done processing this message
    message.ack()

  def run_task(self, client_task):
    """Run a task as commanded by the control server.

    Args:
      client_task: A dictionary of the task to attempt.
    """

    # Start creating the response to send back to the server
    data = {'id': client_task.get('id')}
    task_name = client_task['command']['name']
    routing_key = 'task_complete'

    if task_name == 'refresh':
      # Re-poll available system data
      item = client_task['command'].get('item', 'all')
      self.logger.debug('Retrieving %s information', item)
      info_method = getattr(sysinfo, 'pull_{item}'.format(item=item))

      routing_key = '{user}.{host}.{item}'.format(user=self.user_id,
                                                  host=self.obj_id, item=item)

      data['info'] = info_method()
      data['status'] = 'SUCCESS'
    elif task_name == 'com':
      # Run a command using the Windows COM interface
      if self.current_system == 'Windows':
        interface = client_task['command']['interface']
        item = client_task['command']['item']

        try:
          com_client = wincom.WindowsCOMClient(interface)

          # Check if are retrieving or setting information
          if client_task['command']['action'] == 'set':
            value = client_task['command']['value']
            com_client.set_item(value, item)
          elif client_task['command']['action'] == 'get':
            value = com_client.get_item(item)
            data['info'] = {'item': item, 'value': value}
          else:
            self.logger.error('Unable to perform requested COM action')
            data['status'] = 'NOT_IMPLEMENTED'
        except pythoncom.com_error, error:
          # A com_error likely indicates a bad interface name or item name, so we
          # should mark the task as incompleteable
          data['status'] = 'FAILURE'
          data['info'] = error.strerror
      else:
        self.logger.error('COM interface is only accessible on Windows systems')
        data['status'] = 'FAILURE'
    elif task_name == 'expected_update':
      expected_values = client_task['command']['expected']
      monitor_id = client_task['command']['monitor']
      self.logger.info('Received new %s expected values: %s',
                       monitor_id, expected_values)

      if monitor_id in self.expected_values:
        self.expected_values[monitor_id].update(expected_values)
      else:
        self.logger.warn('Monitor "%s" is not enabled!', monitor_id)
    else:
      self.logger.error('Unable to perform requested task')
      data['status'] = 'NOT_IMPLEMENTED'

    if data['id'] is not None or routing_key != 'task_complete':
      self.publish_info_message(routing_key, json.dumps(data))
      self.logger.info('Sent task completion to server')



#!/usr/bin/env python
"""
Hiveary
https://hiveary.com

Licensed under Simplified BSD License (see LICENSE)
(C) Hiveary, LLC 2013 all rights reserved
"""

import inspect
import json
import logging
import os
import signal
import socket
import sys
from twisted.internet import reactor, task

# esky is only needed for updating the application if its frozen
if hasattr(sys, 'frozen'):
  import esky

# Local imports
from . import __version__
from . import daemon
from . import monitors
from . import network
from . import paths
from . import sysinfo


class RealityAuditor(daemon.Daemon):
  """Daemon subclass. Makes sure that data is being collected and aggregated,
  network connections are active, and gravity is working."""

  INITIAL_DELAY = 5  # Small delay to make sure the network has been initialized
  UPDATE_TIMER = 60 * 60 * 8  # How often to check for agent updates, in seconds
  PID_FILE = '/var/run/hiveary-agent.pid'  # Default location of the PID file
  REMOTE_HOST = 'hiveary.com'  # Default server to connect to
  MONITORS_DIR = '/usr/lib/hiveary/'  # Default location to find monitor modules

  def __init__(self, parsed_args, stored_config, logger=None):
    """Initialization when the agent is started.

    Args:
      parsed_args: A dictionary version of the parsed command line arguments.
      stored_config: A dictionary of any values from the config file.
      logger: Optional, the logger that this class should log to.
    """

    self.logger = logger or logging.getLogger('hiveary_agent.controller')

    # Check that the directory for the pid file exists, and if not then use
    # the same directory as the config file
    pid_file = stored_config.get('pid_file', self.PID_FILE)
    if not os.path.isdir(os.path.dirname(pid_file)):
      directory = os.path.dirname(stored_config['filename'])
      filename = os.path.basename(pid_file)
      pid_file = os.path.join(directory, filename)

    # Check if the monitors directory exists, otherwise, place it under the
    # location of the config file
    self.monitors_dir = stored_config.get('monitors_dir') or self.MONITORS_DIR
    if not os.path.isdir(self.monitors_dir):
      directory = os.path.dirname(stored_config['filename'])
      self.monitors_dir = os.path.join(directory, 'monitors')
      if not os.path.isdir(self.monitors_dir):
        os.makedirs(self.monitors_dir)

    self.monitor_config = stored_config.get('monitors')
    if self.monitor_config is None:
      self.monitor_config = {'resources': ['ResourceMonitor']}
      parsed_args['update'] = True

    self.monitors = []

    # Get and possibly save optional configuration parameters. If the defaults
    # are used, they won't be saved.
    self.extra_options = {}
    for option in ('monitor_backoff', 'pid_file', 'ca_bundle', 'monitors_dir'):
      value = stored_config.get(option)
      if value:
        self.extra_options[option] = value

    # Network controller is initalized by the agent with the necessary
    # authentication credentials
    self.network_controller = network.NetworkController(reactor)
    self.set_config(parsed_args, stored_config)

    # Setup a handler to interpret interrupts since twisted overrides them
    signal.signal(signal.SIGINT, self.signal_handler)

    executable, args = paths.find_executable()
    self.logger.debug('Setting daemon to use %s %s', executable, args)
    super(RealityAuditor, self).__init__(pid_file, executable=executable,
                                         args=args)

  def run(self):
    """Called once the agent has been daemonized, or if the agent is running
    in the foreground. All monitors are started from here and communication
    with the server is started."""

    self.network_controller.ensure_internet_connection()

    if hasattr(sys, 'frozen'):
      # Setup the auto-updater
      update_path = 'https://{host}/versions'.format(host=self.network_controller.remote_host)
      self.app = esky.Esky(sys.executable, update_path)
      reactor.callLater(0, self.start_loop, self.UPDATE_TIMER, True,
                        self.auto_agent_update)

    # Load all monitors
    self.load_monitors()

    # Connect to the server
    self.network_controller.initialize_amqp()

    # Start all of our monitors
    for monitor in self.monitors:
      self.start_monitor(monitor)

    # Send the first data dump
    data = sysinfo.pull_all()
    data['version'] = __version__
    data['host_id'] = self.network_controller.obj_id
    data['monitors'] = []
    for monitor in self.monitors:
      monitor_data = {
          'name': monitor.NAME,
          'sources': monitor.SOURCES,
          'id': monitor.UID,
          'type': monitor.TYPE,
      }
      data['monitors'].append(monitor_data)
    reactor.callLater(self.INITIAL_DELAY,
                      self.network_controller.publish_info_message,
                      'startup',
                      json.dumps(data))

    # Send a ping to the server to act as a keep-alive.
    reactor.callLater(self.INITIAL_DELAY, self.start_loop,
                      self.network_controller.PING_TIMER, True,
                      self.network_controller.ping_pong)

    reactor.run()

  def load_monitors(self):
    """Loads all monitors from the config file. If it cannot find a configured module,
    it will attempt to download one.
    """

    # Check that we have monitors enabled in the config, and know where to find them
    if self.monitor_config and self.monitors_dir:
      # Add the monitors dir to path so we can import monitors
      self.logger.info('Loading monitors from %s', self.monitors_dir)
      sys.path.insert(0, self.monitors_dir)

      # When frozen, the monitors included with the current version also
      # need to be loaded
      if hasattr(sys, 'frozen'):
        frozen_monitors_dir = os.path.join(os.path.dirname(sys.executable),
                                           'monitors')
        sys.path.insert(0, frozen_monitors_dir)
        self.logger.debug('Also loading frozen monitors from: %s',
                          frozen_monitors_dir)

      # Import all of the monitors and add the instances to our monitor list.
      for module_name, monitor_classes in self.monitor_config.iteritems():
        try:
          self.logger.info('Loading monitor %s from file %s', monitor_classes, module_name)
          module = __import__(module_name, globals(), locals(),
                              fromlist=monitor_classes)
          for class_name in monitor_classes:
            monitor_class = getattr(module, class_name)
            # Make sure this class inherits the monitors.BaseMonitor class.
            if monitors.BaseMonitor in inspect.getmro(monitor_class):
              monitor = monitor_class()
              self.monitors.append(monitor)
            else:
              self.logger.warn('Tried to load %s, but was not a HivearyMonitor', monitor_class)
        except:
          self.logger.error('Failed to load module %s', module_name)

  def start_monitor(self, monitor):
    """Starts a given monitor.

    Args:
      monitor: Instance of a monitor class
    """

    self.logger.debug('Starting %s (%s) monitor data checks', monitor.NAME,
                      monitor.UID)
    self.network_controller.expected_values[monitor.UID] = monitor.expected_values
    monitor.send_alert = self.network_controller.publish_alert_message

    # Check if the monitor should run in a loop
    if monitor.MONITOR_TIMER is not None:
      self.start_loop(monitor.MONITOR_TIMER, False, monitor.run)
    else:
      reactor.callInThread(monitor.run)

    self.start_aggregation_loop(monitor)

  def signal_handler(self, signum, stackframe):
    """Handles a SIGTERM or SIGINT sent to the process.

    Args:
      signum: The number of the signal. SIGINT is 2, SIGTERM is 15.
      stackframe: The interrupted stack frame.
    """

    self.logger.info("Received signal: %s", signum)

    self.network_controller.stop_amqp()
    reactor.callFromThread(reactor.stop)

    # Clean up the daemon after the reactor is done
    reactor.addSystemEventTrigger('after', 'shutdown', self.delpid)

  def set_config(self, args, stored_config):
    """Update any configuration variables. Use passed command-line values first,
    then stored config values, then defaults.

    Args:
      args: A dictionary version of the parsed command line arguments.
      stored_config: A dictionary of any values from the config file.
    """

    stored_config.update((k, v) for k, v in args.iteritems() if v is not None)
    self.logger.debug('Using the merged config: %s', stored_config)

    # Store the required network config values in the network controller
    self.network_controller.debug_mode = stored_config['debug']
    self.network_controller.disable_ssl_verification = stored_config['disable_ssl_verify']
    self.network_controller.hostname = socket.getfqdn()
    self.network_controller.owner = stored_config.get('username')
    self.network_controller.access_token = stored_config.get('access_token')
    self.network_controller.remote_host = stored_config.get('server') or self.REMOTE_HOST
    self.network_controller.amqp_server = stored_config.get('amqp_server') or 'amqp.{domain}'.format(
        domain=self.network_controller.remote_host)
    self.network_controller.ca_bundle = stored_config.get('ca_bundle')

    # Update the config file if needed
    if args.get('update'):
      self.update_config_file(stored_config['filename'])

  def update_config_file(self, filename):
    """Stores the current configuration in the config file, overwriting current
    values.

    Args:
      filename: The full path to the config file.
    """

    config = {
        'server': self.network_controller.remote_host,
        'access_token': self.network_controller.access_token,
        'username': self.network_controller.owner,
        'amqp_server': self.network_controller.amqp_server,
        'monitors': self.monitor_config,
    }

    config.update(self.extra_options)

    with open(filename, 'w') as file_desc:
      json.dump(config, file_desc, indent=2)

  def start_aggregation_loop(self, monitor):
    """Starts the aggregation loop, setting it up to use the exact intervals
    if the passed monitor uses intervals.

    Args:
      monitor: The monitor object. Must be a subclass of BaseMonitor.
    """

    if hasattr(monitor, 'next_interval'):
      # Only start aggregating data at a specified time to make sure that we
      # maintain the same time intervals. The first run will contain a partial
      # period if we have data points from at least half that period.
      delta = monitor.next_interval()
      self.logger.debug('Starting %s aggregation loop in %s seconds', monitor.NAME, delta)
      start_now = delta > (monitor.AGGREGATION_TIMER / 2)
    else:
      delta = monitor.AGGREGATION_TIMER
      start_now = True

    reactor.callLater(int(delta), self.start_loop,
                      monitor.AGGREGATION_TIMER, start_now,
                      monitor.send_data, self.network_controller)

  def start_loop(self, timer, start_now, func, *args, **kwargs):
    """Continuously loop the passed function.

    Args:
      timer: How often to loop the call, in seconds.
      start_now: Boolean indicating whether the first iteration should happen
                 immediately or wait until after the first timer expires.
      func: The function to run.
      *args, **kwargs: Anything that needs to be passed to the function.
    """

    loop = task.LoopingCall(func, *args, **kwargs)
    deferred_task = loop.start(timer, now=start_now)
    deferred_task.addErrback(self.logger.error)

  def restart(self):
    """Setup an event to restart the agent after the reactor stops."""

    self.logger.warn('The agent is restarting...')

    reactor.addSystemEventTrigger('after', 'shutdown', self.delpid)
    reactor.addSystemEventTrigger('after', 'shutdown', self.fork,
                                  detached=False, exit=False)

    self.network_controller.stop_amqp()
    reactor.stop()

  def auto_agent_update(self):
    """Checks for a new version of the running application from the remote server.
    If a new version is found, it will be downloaded and extracted, and the agent
    restarted. This only applies if the application is frozen."""

    self.logger.info('Checking for updates, currently running %s',
                     self.app.active_version)

    new_version = self.app.find_update()
    if new_version:
      self.logger.info('Version %s found', new_version)

      self.app.auto_update(callback=self.logger.debug)
      self.logger.info('New version installed')

      self.restart()
    else:
      self.logger.info('No update found')



#!/usr/bin/env python
"""
Hiveary
https://hiveary.com

Licensed under Simplified BSD License (see LICENSE)
(C) Hiveary, LLC 2013 all rights reserved
"""

import atexit
import errno
import os
import platform
import psutil
import signal
import subprocess
import sys
import time

if platform.system() == 'Windows':
  import win32con


class Daemon(object):
  """A generic daemon class, modified from the original.

  Original version: http://www.jejik.com/articles/2007/02/a_simple_unix_linux_daemon_in_python/
  License: https://creativecommons.org/licenses/by-sa/3.0/

  Usage: subclass the Daemon class and override the run() method."""

  def __init__(self, pidfile, term_timeout=20,
               stdin=os.devnull, stdout=os.devnull, stderr=os.devnull,
               executable=sys.executable, args=sys.argv):
    self.stdin = stdin
    self.stdout = stdout
    self.stderr = stderr
    self.pidfile = pidfile
    self.term_timeout = term_timeout
    self.exec_params = [executable] + args

  def fork(self, detached=True, exit=True):
    """Do the UNIX double-fork magic, see Stevens' "Advanced
    Programming in the UNIX Environment" for details (ISBN 0201563177)
    http://www.erlenstar.demon.co.uk/unix/faq_2.html#SEC16
    On Windows we can just create a detached subprocess.

    Args:
      detached: Whether the new process should be started completely independantly,
          or hook into the current stdin/stderr/stdout.
      exit: Whether to exit after forking. This should be true unless the process
          will exit somewhere else.
    """

    if platform.system() == 'Windows':
      # Clean out the params so that the forked process starts normally, without
      # trying to daemonize itself again
      if 'start' in self.exec_params:
        self.exec_params.remove('start')
      if 'restart' in self.exec_params:
        self.exec_params.remove('restart')

      if detached:
        proc = subprocess.Popen(self.exec_params,
                                creationflags=win32con.DETACHED_PROCESS,
                                close_fds=True)
      else:
        proc = subprocess.Popen(self.exec_params)

      # write pidfile and exit
      self.write_pid(proc.pid)
      if exit:
        sys.exit()
    else:
      try:
        pid = os.fork()
        if pid > 0:
          # exit first parent
          sys.exit(0)
      except OSError, e:
        sys.stderr.write("fork #1 failed: %d (%s)\n" % (e.errno, e.strerror))
        sys.exit(1)

      # decouple from parent environment
      os.chdir("/")
      os.setsid()
      os.umask(0)

      # do second fork
      try:
        pid = os.fork()
        if pid > 0:
          # exit from second parent
          sys.exit(0)
      except OSError, e:
        sys.stderr.write("fork #2 failed: %d (%s)\n" % (e.errno, e.strerror))
        sys.exit(1)

      # redirect standard file descriptors
      sys.stdout.flush()
      sys.stderr.flush()
      si = file(self.stdin, 'r')
      so = file(self.stdout, 'a+')
      se = file(self.stderr, 'a+', 0)
      os.dup2(si.fileno(), sys.stdin.fileno())
      os.dup2(so.fileno(), sys.stdout.fileno())
      os.dup2(se.fileno(), sys.stderr.fileno())

      # write pidfile
      atexit.register(self.delpid)
      self.write_pid(str(os.getpid()))

  def delpid(self):
    """Deletes the PID file for the daemon if it exists."""

    if os.path.exists(self.pidfile):
      os.remove(self.pidfile)

  def start(self):
    """Start the daemon."""

    # Check for a pidfile to see if the daemon already runs
    try:
      with open(self.pidfile, 'r') as pf:
        pid = int(pf.read().strip())
    except IOError:
      pid = None

    if pid:
      message = "pidfile %s already exist. Daemon already running?\n"
      sys.stderr.write(message % self.pidfile)
      sys.exit(1)

    # Start the daemon
    self.fork(detached=True, exit=True)
    self.run()

  def stop(self):
    """Stop the daemon."""

    pid = self.get_pid()
    start_time = time.time()
    timeout = start_time + self.term_timeout

    if not pid:
      message = "pidfile %s does not exist. Daemon not running?\n"
      sys.stderr.write(message % self.pidfile)
      return  # not an error in a restart

    # Try gently stopping the daemon, and escalate if it does not exit
    try:
      while time.time() < timeout:
        os.kill(pid, signal.SIGINT)
        time.sleep(0.5)

      if hasattr(signal, 'SIGKILL'):
        os.kill(pid, signal.SIGKILL)
        self.delpid()
      else:
        message = 'Failed to terminate process %s\n' % pid
        sys.stderr.write(message)
        sys.exit(1)
    except OSError, err:
      if err.errno != errno.EPERM and err.errno != errno.EACCES:
        self.delpid()
      else:
        raise

  def get_pid(self):
    """Get the pid from the pidfile."""

    try:
      with open(self.pidfile, 'r') as pf:
        pid = int(pf.read().strip())
    except IOError:
      pid = None

    return pid

  def status(self):
    """Output the status of the daemon."""

    pid = self.get_pid()
    if pid is None:
      message = 'Hiveary Agent is not running\n'
      sys.stdout.write(message)
      sys.exit(1)

    # Check for the existence of a process with the pid
    pid_exists = psutil.pid_exists(pid)
    if not pid_exists:
      message = 'Pidfile contains pid %s, but no running process could be found\n' % pid
      sys.stderr.write(message)
      sys.exit(1)

    message = 'Hiveary Agent is running with pid %s\n' % pid
    sys.stdout.write(message)
    sys.exit()

  def write_pid(self, pid):
    """Updates the pid in the pidfile."""

    with file(self.pidfile, 'w') as pf:
      pf.write('%s\n' % pid)

  def run(self):
    """You should override this method when you subclass Daemon. It will be
    called after the process has been daemonized by start() or restart().

    Raises:
      NotImplementedError: The method hasn't been overridden in the subclass
                           and will never do anything.
    """

    raise NotImplementedError



#!/usr/bin/env python
"""
Hiveary
https://hiveary.com

Licensed under Simplified BSD License (see LICENSE)
(C) Hiveary, LLC 2013 all rights reserved
"""

import httplib2
import logging
from oauth2 import Consumer, Client, Request, Token, SignatureMethod_HMAC_SHA1
import os
import urllib
from urlparse import parse_qs

# Local imports
from . import paths


class OAuthClient(Client):
  """Subclass of the oauth Client that allows the use of header auth."""

  def __init__(self, consumer, token=None, cache=None, timeout=None,
               proxy_info=None, debug=False, disable_ssl_verification=False,
               ca_bundle=None, logger=None):
    """Initialize the client.

    Args:
      consumer: An instance of oauth2.Consumer, used for 2-legged OAuth requests
      token: An instance of oauth2.Token, used for 2-legged OAuth requests
      cache: If 'cache' is a string then it is used as a directory name for
             a disk cache. Otherwise it must be an object that supports the
             same interface as httplib2.FileCache.
      timeout: Timeout in seconds. If None is passed for timeout
               then Python's default timeout for sockets will be used. See
               for example the docs of socket.setdefaulttimeout():
               http://docs.python.org/library/socket.html#socket.setdefaulttimeout
      proxy_info: May be:
                  - a callable that takes the http scheme ('http' or 'https') and
                    returns a httplib2.ProxyInfo instance per request.
                    By default, uses proxy_nfo_from_environment.
                  - a httplib2.ProxyInfo instance (static proxy config).
                  - None (proxy disabled).
      debug: Boolean indicating whether the agent is running in debug mode.
      disable_ssl_verification: Boolean indicating whether the server's SSL
                                certificate should be verified.
      ca_bundle: The full path to the CA bundle used for SSL verification.
      logger: An existing logging object to use.
    Raises:
      IOError: The CA bundle was not found on the system, and SSL verification
               is enabled.
    """

    self.logger = logger or logging.getLogger('hiveary_agent.oauth')

    if consumer is not None and not isinstance(consumer, Consumer):
        raise ValueError("Invalid consumer.")

    if token is not None and not isinstance(token, Token):
        raise ValueError("Invalid token.")

    self.consumer = consumer
    self.token = token
    self.method = SignatureMethod_HMAC_SHA1()

    if ca_bundle is None or not os.path.isfile(ca_bundle):
      ca_bundle = os.path.join(paths.get_program_path(), 'ca-bundle.pem')

    self.logger.debug('Using %s as the CA bundle', ca_bundle)

    if disable_ssl_verification:
      self.logger.warn('SSL validation is disabled')
    elif not os.path.isfile(ca_bundle):
      self.logger.error('CA bundle %s does not exist!', ca_bundle)
      raise IOError('File does not exist')

    httplib2.Http.__init__(self, cache=cache, timeout=timeout,
                           proxy_info=proxy_info, ca_certs=ca_bundle,
                           disable_ssl_certificate_validation=disable_ssl_verification)

  def request(self, uri, method="GET", body='', headers=None,
              redirections=httplib2.DEFAULT_MAX_REDIRECTS, connection_type=None):
    """Creates and sends a two-legged OAuth 1.0a authenticated request.

    Args:
      uri: The absolute uri of the server.
      method: The HTTP method to use.
      body: Any body that should go with the request, as a string.
      headers: A dictionary of any extra heders to include.
      redirection: Passed directly to httplib2
      connection_type: Passed directly to httplib2
    Returns:
      A tuple of (response, content), the first being an instance of the
      'httplib2.Http.Response' class, the second being a string that
      contains the response entity body.
    """

    DEFAULT_POST_CONTENT_TYPE = 'application/x-www-form-urlencoded'

    # Body must be a string for the OAuth signature
    if body is None:
      body = ''

    if not isinstance(headers, dict):
      headers = {}

    if method == "POST":
      headers['Content-Type'] = headers.get('Content-Type',
                                            DEFAULT_POST_CONTENT_TYPE)

    is_form_encoded = headers.get('Content-Type') == 'application/x-www-form-urlencoded'

    if is_form_encoded and body:
      parameters = parse_qs(body)
    else:
      parameters = None

    req = Request.from_consumer_and_token(
        self.consumer, token=self.token, http_method=method, http_url=uri,
        parameters=parameters, body=body, is_form_encoded=is_form_encoded)

    req.sign_request(self.method, self.consumer, self.token)

    schema, rest = urllib.splittype(uri)
    if rest.startswith('//'):
      hierpart = '//'
    else:
      hierpart = ''
    host, rest = urllib.splithost(rest)

    realm = schema + ':' + hierpart + host

    if is_form_encoded:
      body = req.to_postdata()

    headers.update(req.to_header(realm=realm))

    self.logger.info('Sending %s %s', method, uri)

    return httplib2.Http.request(
        self, uri, method=method, body=body,
        headers=headers, redirections=redirections,
        connection_type=connection_type)



#!/usr/bin/env python
"""
Hiveary
https://hiveary.com

Licensed under Simplified BSD License (see LICENSE)
(C) Hiveary, LLC 2013 all rights reserved

Base classes for collecting data.
"""

from collections import defaultdict
import copy
import datetime
import json
import logging
import time


class BaseMonitor(object):
  """Default class for defining a monitor. This does nothing on its own, it should
  be subclassed with at least get_data defined."""

  AGGREGATION_TIMER = 1800  # 30 minutes
  FLOP_PROTECTION_COUNTER = 6
  MONITOR_TIMER = 1
  DEFAULT_ALERT_BACKOFF = 3600  # Delay between similar alerts, in seconds
  NAME = 'base'
  TYPE = None
  UID = None  # Can be set to any value guaranteed to be unique, a uuid.uuid4() is recommended
  SOURCES = {}

  def __init__(self, backoff=None, logger=None):
    """Initialize the monitor.

    Args:
      backoff: The amount of time to delay between repeated alerts, in seconds.
      logger: A logging object to use.
    Raise:
      AttributeError: The monitor subclass did not set a unique ID.
    """

    self.logger = logger or logging.getLogger('hiveary_agent.%s' % self.NAME)

    if self.UID is None:
      raise AttributeError('UID is not set for monitor %s!' % self.NAME)

    self.logger.info('Monitoring the following sources: %s', self.SOURCES)

    self.backoff = backoff or self.DEFAULT_ALERT_BACKOFF
    self.logger.info('Using an alert backoff of %d seconds', self.backoff)

    self.expected_values = {}
    self.data_points = []

    self.send_alert = None
    self.alert_counters = defaultdict(lambda: 0)
    self.alert_delays = {}

    # Store just the source names separately to prevent having to repeatedly
    # look them up
    self.monitored_source_names = set(self.SOURCES.keys())

  def store_data_point(self, data):
    """Puts a single data point into the database.

    Args:
      data: A dictionary of any data to be stored.
    Raises:
      AttributeError: The data reported by the monitor does not match what it
          declared it would be monitoring.
    """

    monitor_data = copy.copy(data)
    monitor_data.pop('extra', None)

    if set(monitor_data.keys()) != self.monitored_source_names:
      raise AttributeError('The returned data does not match the '
                           'source list for monitor %s' % self.NAME)

    monitor_data['timestamp'] = time.time()
    self.data_points.append(monitor_data)

  def send_data(self, net_controller):
    """Sends the usage data points for the past time period.

    Args:
      net_controller: A NetworkController object with an active AMQP connection.
    """

    # Determine how much stored data to send based on the defined monitor periods
    now_dt = datetime.datetime.utcnow()
    earliest = time.time() - datetime.timedelta(minutes=30,
                                                seconds=now_dt.second,
                                                microseconds=now_dt.microsecond).total_seconds()
    seconds = (now_dt - now_dt.min).seconds
    rounding = (seconds + self.AGGREGATION_TIMER/2) // self.AGGREGATION_TIMER * self.AGGREGATION_TIMER
    time_period = now_dt + datetime.timedelta(0, rounding-seconds, -now_dt.microsecond)

    data = defaultdict(list)
    data['period'] = time_period.strftime('%H%M')
    data['day'] = time_period.weekday()
    data['host_id'] = net_controller.obj_id
    data['id'] = self.UID

    for point in self.data_points:
      # Throw out any data points that are too old
      timestamp = point.pop('timestamp')
      if timestamp >= earliest:
        for source, value in point.iteritems():
          data[source].append(value)

    self.logger.debug('Full data since %s: %s', earliest, data)

    # Send the full data up to the server.
    net_controller.publish_info_message(self.TYPE, json.dumps(data))
    self.data_points = []

  def run(self):
    """Wrapper call to get the data for monitored sources and check it against
    any set alert values."""

    pass

  def alert_check(self, data):
    """Checks the monitored data to determine if an alert should be created.
    This should be defined by a child class for alerting.

    Args:
      data: A dictionary of the data to check. The keys must match the list of
          monitored sources for the class.
    """

    pass

  def get_data(self):
    """Retrieves the monitored data. This must be defined by a child class.

    Raises:
      NotImplementedError: The method hasn't been overridden in the subclass
                           and will never do anything.
    """

    raise NotImplementedError

  def extra_alert_data(self, source, data):
    """Finds additional information that should be sent when an alert is fired
    for this monitor. This should be defined by a child class.

    Args:
      source: The source of the fired alert.
      data: A dictionary of any additional information generated from the monitor
          while checking its sources.
    """

    pass


class PollingMixin(object):
  """Mixin class for monitors that expect to regularly poll their data sources
  for new data."""

  def run(self):
    """Wrapper call to get the data for monitored sources and check it against
    any set alert values."""

    data = self.get_data()
    self.store_data_point(data)
    self.alert_check(data)


class IntervalMixin(object):
  """Mixin class for handling regular interval-based aggregation -
  i.e. send data exactly on the hour."""

  def next_interval(self, now=None, interval=None):
    """Returns the time in seconds until the next interval.

    Args:
      now: The base-time for determining the next interval.
      interval: Amount of time in between aggreagtion intervals, in seconds.
    Returns:
      An integer indicating the time in seconds until the next interval.
    """

    if now is None:
      now = datetime.datetime.utcnow()
    if interval is None:
      interval = self.AGGREGATION_TIMER

    nsecs_passed = now.minute * 60 + now.second + now.microsecond * 1e-6
    delta = (nsecs_passed // interval) * interval + interval - nsecs_passed

    return delta


class UsageMonitor(IntervalMixin, BaseMonitor):
  """Base class for all "usage" type monitors."""

  TYPE = 'usage'

  def alert_check(self, data):
    """Checks the monitored data to determine if an alert should be created.

    Args:
      data: A dictionary of the data to check. The keys must match the list of
          monitored sources for the class.
    """

    now = time.time()

    for source in self.monitored_source_names:
      delay = self.alert_delays.get(source)
      threshold = self.expected_values.get(source)
      usage = data.get(source)

      if delay and delay <= now:
        self.alert_delays.pop(delay)
        delay = None

      if threshold and not delay and usage >= threshold:
        self.logger.debug('Current %s usage at %s, threshold is %s', source,
                          usage, threshold)

        self.alert_counters[source] += 1
        if self.alert_counters[source] >= self.FLOP_PROTECTION_COUNTER:
          # Send an alert to the server with any extra information for this source
          extra = data.pop('extra', {})
          alert = {
              'threshold': threshold,
              'current_usage': usage,
              'timestamp': now,
              'monitor': {
                  'id': self.UID,
                  'name': self.NAME,
                  'type': self.TYPE,
                  'source': source,
                  'source_type': self.SOURCES[source],
              },
              'event_data': self.extra_alert_data(source, extra) or {},
          }
          self.send_alert(alert)

          # Put a delay on the next alert to prevent a flood of alert messages
          self.alert_delays[source] = now + self.backoff
          self.alert_counters[source]
      else:
        self.alert_counters[source]


class LogMonitor(BaseMonitor):
  """Base class for all "log" type monitors."""

  TYPE = 'log'


class StatusMonitor(BaseMonitor):
  """Base class for all "status" type monitors."""

  TYPE = 'status'



#!/usr/bin/env python
"""
Hiveary
https://hiveary.com

Licensed under Simplified BSD License (see LICENSE)
(C) Hiveary, LLC 2013 all rights reserved

Hiveary Resource Monitor
Monitors the following sources:
  bytes_sent, bytes_recv, disk, cpu, ram
"""

import copy
import psutil
import time

from hiveary import monitors, sysinfo


class ResourceMonitor(monitors.PollingMixin, monitors.UsageMonitor):
  """Monitors system resource data."""

  MONITOR_TIMER = 10
  NAME = 'resources'
  UID = '2c72af48-37ce-4ea1-9e53-9f081a6bcb6b'

  def __init__(self, *args, **kwargs):
    # Expected resource usage parameters for the current time frame, stored as percentages
    self.disks = sysinfo.find_valid_disks()

    self.SOURCES = {
        'ram': 'percent',
        'cpu': 'percent',
        'bytes_sent': 'bytes',
        'bytes_recv': 'bytes',
    }
    for disk in self.disks:
      self.SOURCES['disk_' + disk] = 'percent'

    # Initialize the network information
    self.total_net_io = psutil.network_io_counters()
    self.last_check = time.time()

    super(ResourceMonitor, self).__init__(*args, **kwargs)

  def get_data(self):
    """Gets the system's resource usage and sends an alert when it becomes too high."""

    now = time.time()

    # We have to aggregate usage so we'll pull everything regardless of the params
    network_io = psutil.network_io_counters()
    time_diff = now - self.last_check
    ram_usage = psutil.phymem_usage()
    disk_usage = {}
    for disk in self.disks:
      disk_usage[disk] = psutil.disk_usage(disk)

    current_usage = {
        'bytes_sent': (network_io.bytes_sent - self.total_net_io.bytes_sent) / time_diff,
        'bytes_recv': (network_io.bytes_recv - self.total_net_io.bytes_recv) / time_diff,
        'ram': ram_usage.percent,
        'cpu': psutil.cpu_percent(interval=None),
    }

    extra_data = {
        'ram': {
            'total_memory': ram_usage.total,
            'used_memory': ram_usage.used,
            'free_memory': ram_usage.free,
            'resource': 'RAM',
        },
        'cpu': {
            'resource': 'CPU',
        }
    }

    # Add in disk usage data
    for device, usage in disk_usage.iteritems():
      disk_name = 'disk_%s' % device
      current_usage[disk_name] = usage.percent
      extra_data[disk_name] = {
          'disk': device,
          'total_space': usage.total,
          'used_space': usage.used,
          'free_space': usage.free,
      }

    self.last_check = now
    self.total_net_io = network_io

    # Store the values
    current_usage['extra'] = extra_data
    return current_usage

  def extra_alert_data(self, source, data):
    """Finds additional information that should be sent when an alert is fired
    for this monitor.

    Args:
      source: The source of the fired alert.
      data: A dictionary of information specific to the monitored resources.
    Returns:
      A dictionary containing the additonal information to send with the alert.
    """

    if source == 'ram':
      top = 'memory_percent'
    elif source == 'cpu':
      top = 'cpu_percent'
    else:
      top = None

    procs, top_procs = sysinfo.pull_processes(top=top)
    extra_data = copy.copy(data.get(source, {}))
    extra_data['current_procesess'] = procs

    if top_procs:
      extra_data['top_processes'] = top_procs

    return extra_data



#!/usr/bin/env python
"""
Hiveary
https://hiveary.com

Licensed under Simplified BSD License (see LICENSE)
(C) Hiveary, LLC 2013 all rights reserved

Functions for collecting local system information.
"""

import logging
import netifaces
import os
import platform
import psutil
import simplejson
import socket
import subprocess

if subprocess.mswindows:
  import win32net
  import win32service

  import wincom
else:
  import grp
  import pwd


logger = logging.getLogger('hiveary_agent.sysinfo')
current_system = platform.system()


def find_valid_disks():
  """Iterate through and find all valid disk partitions. Some devices will not
  be accessible since they have non-disk filesystems, such as a CD-ROM drive.

  Returns:
    A list of all valid disks.
  """

  disks = []
  for disk in psutil.disk_partitions():
    try:
      psutil.disk_usage(disk.device)
    except OSError:
      continue
    else:
      disks.append(disk.device)

  return disks


def pull_all():
  """Finds all available static information about the host.

  Returns:
    A dictionary containing all discovered information.
  """

  logger.info('Pulling all local info')
  info = {}

  info['os'], info['processor'] = pull_os()
  info['disks'] = pull_disks()
  info['fqdn'], info['interfaces'] = pull_net()
  info['users'] = pull_users()
  info['groups'] = pull_groups()
  info['services'] = pull_services()
  info['automatic_update_settings'] = pull_update_settings()

  return info


def pull_os():
  """Find information about the host's operating system.

  Returns:
    A tuple of a dictionary of the OS information, and the system's processor.

    Example:
    (
     {
       'architecture': 'nt',
        'platform': 'Windows',
        'release': {
          'product_name': 'Windows 8',
          'service_pack': 0,
          'version_number': '6.2'
      },
     'Intel64 Family 6 Model 58 Stepping 9, GenuineIntel'
    )
  """

  os_info = {}
  os_info['architecture'] = os.name
  os_info['platform'] = current_system
  if current_system == 'Windows':
    os_info['release'] = wincom.get_version_info()
  else:
    os_info['release'] = platform.release()

  processor = platform.processor()

  return (os_info, processor)


def pull_disks():
  """Finds full disk information from the host.

  Returns:
    A dictionary of all disk-like devices mapped to their data. Example:

    {
      'C:\\': {'filesystem': 'NTFS', 'mountpoint': 'C:\\', 'options': 'rw,fixed'},
      'D:\\': {'filesystem': 'NTFS', 'mountpoint': 'D:\\', 'options': 'rw,fixed'},
      'E:\\': {'filesystem': 'NTFS', 'mountpoint': 'E:\\', 'options': 'rw,fixed'},
      'H:\\': {'filesystem': '', 'mountpoint': 'H:\\', 'options': 'cdrom'}
    }
  """

  disks = {}
  for disk in psutil.disk_partitions():
    disks[disk.device] = {
        'mountpoint': disk.mountpoint,
        'filesystem': disk.fstype,
        'options': disk.opts,
        }
  return disks


def pull_net():
  """Finds network related information.

  Returns:
    A tuple of the host's FQDN and a list of network interfaces. Each network
    interface in the list is a dictionary of the interface properties. Example:

    (
      'prism7.utah.nsa.gov',
      [
        {
        'af_inet': {
          'addr': '127.0.0.1',
          'netmask': '255.0.0.0',
          'peer': '127.0.0.1'
        },
        'af_inet6': {
          'addr': '::1',
          'netmask': 'ffff:ffff:ffff:ffff:ffff:ffff:ffff:ffff'},
          'af_packet': {
            'addr': '00:00:00:00:00:00',
            'peer': '00:00:00:00:00:00'
          },
        'name': 'lo'
        },
        ...
      ]
    )
  """

  fqdn = socket.getfqdn()
  interfaces = []
  for iface in netifaces.interfaces():
    ifaddrs = netifaces.ifaddresses(iface)
    interface = {'name': iface}

    # Translate the address family identifier to a name
    for num, addrs in ifaddrs.iteritems():
      family = netifaces.address_families[num]

      # Parse out the netmask for IPv6
      if family == 'AF_INET6':
        addresses = []
        for address in addrs:
          addr = address.pop('addr')
          split_addr = addr.split('%')
          address['addr'] = split_addr[0]

          if len(split_addr) == 2:
            address['scope_id'] = split_addr[1]

          addresses.append(address)
      else:
        addresses = addrs

      if len(addresses) == 1:
        addresses = addresses[0]
      interface[family.lower()] = addresses

    interfaces.append(interface)

  logger.debug('Retrieved info on %s interfaces', len(interfaces))
  return (fqdn, interfaces)


def pull_users():
  """Finds local user information.

  Returns:
    A list of users. Each user is a dictionary of user data. Example:

    [
      {
        'gecos': 'root',
        'gid': 0,
        'homedir': '/root',
        'name': 'root',
        'shell': '/bin/bash',
        'uid': 0
      },
      ...
    ]
  """

  if current_system == 'Windows':
    net_users = win32net.NetUserEnum(None, 3)
    logger.debug('Found %s users', len(net_users[0]))
    resume = net_users[2]

    for user in net_users[0]:
      del(user['logon_hours'])  # Hex isn't JSONable
    users = net_users[0]

    while resume != 0:
      net_users = win32net.NetUserEnum(None, 3, resumeHandle=resume)
      logger.debug('Found %s users', len(net_users[0]))
      resume = net_users[2]

      for user in net_users[0]:
        del(user['logon_hours'])  # Hex isn't JSONable
      users.append(net_users[0])
  elif current_system == 'Linux' or current_system == 'Darwin':
    # pwd/grp work on linux and osx
    users = []
    users_pwd = pwd.getpwall()
    for user_pwd in users_pwd:
      user = {
          'name': user_pwd.pw_name,
          'uid': user_pwd.pw_uid,
          'gid': user_pwd.pw_gid,
          'gecos': user_pwd.pw_gecos,
          'homedir': user_pwd.pw_dir,
          'shell': user_pwd.pw_shell
        }
      users.append(user)
  else:
    logger.error('Unkown platform "%s", cannot pull users', current_system)
    return []

  return users


def pull_groups():
  """Finds local group information.

  Returns:
    A list of groups. Each group is a dictionary of information about the group.

    Example:
    [
      {
        'gid': 0,
        'members': ['root'],
        'name': 'root'
      },
      ...
    ]
  """

  if current_system == 'Windows':
    net_local_groups = win32net.NetLocalGroupEnum(None, 1)
    logger.debug('Found %s groups', len(net_local_groups[0]))
    resume = net_local_groups[2]
    groups = net_local_groups[0]

    while resume != 0:
      net_local_groups = win32net.NetLocalGroupEnum(None, 1, resumeHandle=resume)
      logger.debug('Found %s groups', len(net_local_groups[0]))
      resume = net_local_groups[2]
      groups.extend(net_local_groups[0])

    # Enumerate group members
    for group in groups:
      net_members = win32net.NetLocalGroupGetMembers(None, group['name'], 1)
      resume = net_members[2]
      group['members'] = [x['name'] for x in net_members[0]]

      while resume != 0:
        net_members = win32net.NetLocalGroupGetMembers(None, group['name'], 1, resumeHandle=resume)
        resume = net_members[2]
        group['members'].extend([x['name'] for x in net_members[0]])
  elif current_system == 'Linux' or current_system == 'Darwin':
    groups = []
    groups_grp = grp.getgrall()
    for group_grp in groups_grp:
      group = {
          'name': group_grp.gr_name,
          'gid': group_grp.gr_gid,
          'members': group_grp.gr_mem
        }
      groups.append(group)
  else:
    logger.error('Unkown platform "%s", cannot pull groups', current_system)
    return []

  return groups


def pull_services():
  """Finds information about installed services, currently only applicable to
  Windows.

  Returns:
    A dictionary of service name mapped to service information. All strings will
    be in unicode. The service information primarily includes the user-displayed
    service name and a dictionary of status information. Example:

    {
      u'wuauserv': {
        'display_name': u'Windows Update',
        'status': {
          'checkpoint': 0,
          'controls_accepted': 0,
          'service_exit_code': 0,
          'state': 1,
          'type': 32,
          'wait_hint': 0,
          'win32_exit_code': 0
        }
      },
      ...
    }
  """

  if current_system == 'Windows':
    services = {}
    scm = win32service.OpenSCManager(
        None, None, win32service.SC_MANAGER_ENUMERATE_SERVICE)

    raw_services = win32service.EnumServicesStatus(
        scm, win32service.SERVICE_WIN32)
    scm.Close()

    # Parse the services into something useful
    for raw_service in raw_services:
      service = {}
      service['display_name'] = raw_service[1]

      # Parse the status flags into meaningful names
      raw_status = raw_service[2]
      status = {
          'type': raw_status[0],
          'state': raw_status[1],
          'controls_accepted': raw_status[2],
          'win32_exit_code': raw_status[3],
          'service_exit_code': raw_status[4],
          'checkpoint': raw_status[5],
          'wait_hint': raw_status[6]
        }

      service['status'] = status

      # Key the service by ServiceName
      services[raw_service[0]] = service
  else:
    # TODO: nix services and osx services
    logger.warn('Service enumeration not implemented for %s', current_system)
    return {}

  logger.debug('Retrieved the services info')
  return services


def pull_processes(top=None, top_number=5):
  """Retrieves information about active processes.

  Args:
    top: If provided, the `top_number` of processes for the matching
        resource will be returned. Valid options are memory_percent or
        cpu_percent
    top_number: Number of top processes to return for the given resource. Only
        valid when top is not None.
  Returns:
    Returns a tuple of (all processes, top processes).
    Each process in the list is a dictionary with nicode keys containing
    a large amount of detailed information for the process. Example:

    [
      {
        u'cmdline': [],
        u'connections': [
          {
            u'family': 23,
            u'fd': -1,
            u'laddr': [u'::1', 8888],
            u'raddr': [u'::1', 52912],
            u'status': u'TIME_WAIT',
            u'type': 1
          },
          ...
        ],
        u'cpu_affinity': None,
        u'cpu_percent': 0.0,
        u'cpu_times': {u'system': 526099.53125, u'user': 0.0},
        u'create_time': 1381682351.719,
        u'cwd': None,
        u'exe': None,
        u'ext_memory_info': {
          u'nonpaged_pool': 0,
          u'num_page_faults': 0,
          u'paged_pool': 0,
          u'pagefile': 0,
          u'peak_nonpaged_pool': 0,
          u'peak_paged_pool': 0,
          u'peak_pagefile': 0,
          u'peak_wset': 0,
          u'private': 0,
          u'wset': 20480
        },
        u'io_counters': {
          u'read_bytes': 0,
          u'read_count': 0,
          u'write_bytes': 0,
          u'write_count': 0
        },
        u'ionice': None,
        u'memory_info': {u'rss': 20480, u'vms': 0},
        u'memory_maps': None,
        u'memory_percent': 0.00023943963463427274,
        u'name': u'System Idle Process',
        u'nice': None,
        u'num_ctx_switches': {u'involuntary': 0, u'voluntary': 703011786},
        u'num_handles': 0,
        u'num_threads': 4,
        u'open_files': [],
        u'pid': 0,
        u'ppid': 0,
        u'status': u'running',
        u'threads': None,
        u'username': u'NT AUTHORITY\\SYSTEM'
      },
      ...
    ]
  """

  # Create a JSONable version of the process list
  processes = []
  for p in psutil.process_iter():
    try:
      processes.append(simplejson.loads(simplejson.dumps(p.as_dict())))
    except psutil.NoSuchProcess:
      # Likely poor timing for a terminating process, not worth logging
      continue
    except OSError:
      if p.name == 'System':
        # System cannot be converted to a dict, just ignore it
        continue
      else:
        raise

  # Find the top processes
  top_procs = []
  if top:
    full_top_procs = sorted(processes, key=lambda p: p[top], reverse=True)
    top_procs = []

    for i in xrange(0, top_number):
      try:
        proc = full_top_procs[i]
      except IndexError:
        break

      # Pull out just a subset of information
      proc_subset = {
          'name': proc['name'],
          'pid': proc['pid'],
          top: proc[top],
      }
      top_procs.append(proc_subset)

    logger.debug('Top processes for %s: %s', top, top_procs)

  logger.debug('Retrieved the running processes')
  return processes, top_procs


def pull_update_settings():
  """Retrieves information about the system's auto update settings, currently
  only applicable to Windows.

  Retruns:
    A dictionary of the update settings. Example:

    {
      'service_enabled': True,
      'notification_level': 'Scheduled Installation',
      'installation_schedule': {'day': 'Every Day', 'time': '03:00'}
    }
  """

  logger.debug('Retrieving system update settings.')

  if current_system == 'Windows':
    automatic_update_settings = wincom.get_update_settings()
    logger.debug('Retrieved the following update settings: %s',
                 automatic_update_settings)
  else:
    automatic_update_settings = {}

  return automatic_update_settings



#!/usr/bin/env python
"""
Hiveary
https://hiveary.com

Licensed under Simplified BSD License (see LICENSE)
(C) Hiveary, LLC 2013 all rights reserved

Module containing functions related to the Windows COM interface.
"""

import subprocess

# OS specific imports
if subprocess.mswindows:
  import pythoncom
  import win32api
  import win32com.client


# Constants
AUTO_UPDATE_SETTINGS = {
    'InstallationTime': [
        '00:00',
        '01:00',
        '02:00',
        '03:00',
        '04:00',
        '05:00',
        '06:00',
        '07:00',
        '08:00',
        '09:00',
        '10:00',
        '11:00',
        '12:00',
        '13:00',
        '14:00',
        '15:00',
        '16:00',
        '17:00',
        '18:00',
        '19:00',
        '20:00',
        '21:00',
        '22:00',
        '23:00'
      ],
    'InstallationDay': [
        'Every Day',
        'Every Sunday',
        'Every Monday',
        'Every Tuesday',
        'Every Wednesday',
        'Every Thursday',
        'Every Friday',
        'Every Saturday'
      ],
    'NotificationLevel': [
        'Not Configured',
        'Disabled',
        'Notify Before Download',
        'Notify Before Installation',
        'Scheduled Installation'
      ]
  }

WINDOWS_VERSIONS = {
    5: [
        ['', 'Windows 2000 Professional', 'Windows 2000 Server', 'Windows 2000 Server'],
        ['', 'Windows XP', '', ''],
        ['', 'Windows XP 64-bit', 'Windows Server 2003', 'Windows Server 2003'],
      ],
    6: [
        ['', 'Windows Vista', 'Windows Server 2008', 'Windows Server 2008'],
        ['', 'Windows 7', 'Windows Server 2008 R2', 'Windows Server 2008 R2'],
        ['', 'Windows 8', 'Windows Server 2012', 'Windows Server 2012']
      ]
  }


class WindowsCOMClient(object):
  """Class for executing COM commands."""

  def __init__(self, name):
    """Initialize the class.

    Args:
      name: The full dot-separated name of the COM object (ie Excel.Application)
    """

    pythoncom.CoInitialize()
    self.client = win32com.client.Dispatch(name)

  def get_item(self, hierarchy):
    """Retrieve the item as specified by the hierarchy ordering.

    Args:
      hierarchy: A list or dot-separated string of the specific item to retrieve.

    Returns:
      The retrieved COM item.
    """

    if type(hierarchy) == str or type(hierarchy) == unicode:
      hierarchy = hierarchy.split('.')

    item = self.client
    for level in hierarchy:
      item = getattr(item, level)

    return item

  def set_item(self, value, hierarchy):
    """Set the item as specified by the hierarchy ordering to the passed value.

    Args:
      value: The new value for the COM object.
      hierarchy: A list or dot-separated string of the specific item to retrieve.
    """

    if type(hierarchy) == str or type(hierarchy) == unicode:
      hierarchy = hierarchy.split('.')

    item = self.client
    for level in hierarchy[:-1]:
      item = getattr(item, level)

    setattr(item, hierarchy[-1], value)


def get_version_info():
  """Returns the operating system version info.

  Returns:
    A dictionary of the operating system version info. Example:

    {
      'product_name': 'Windows 8',
      'service_pack': 0,
      'version_number': '6.2'
    }
  """

  version_tuple = win32api.GetVersionEx(1)
  # 0 is the major version, 1 is the minor version, and 8 is the product type
  # Product type can be:
  #   1 (VER_NT_WORKSTATION) - workstation
  #   2 (VER_NT_SERVER) - server
  #   3 (VER_NT_DOMAIN_CONTROLLER) - server and domain controller
  version_name = WINDOWS_VERSIONS[version_tuple[0]][version_tuple[1]][version_tuple[8]]

  version_info = {
      'product_name': version_name,
      'version_number': '%s.%s' % version_tuple[:2],
      'service_pack': version_tuple[5]
    }

  return version_info


def get_update_settings():
  """Returns the current settings related to Windows updates.

  Returns:
    A dictionary of the current Windows update settings. Example:

    {
      'service_enabled': True,
      'notification_level': 'Scheduled Installation',
      'installation_schedule': {'day': 'Every Day', 'time': '03:00'}
    }
  """

  auto_client = WindowsCOMClient('Microsoft.Update.AutoUpdate')

  notification_level_code = auto_client.get_item('Settings.NotificationLevel')
  notification_level = AUTO_UPDATE_SETTINGS['NotificationLevel'][notification_level_code]

  settings = {
      'service_enabled': auto_client.get_item('ServiceEnabled'),
      'notification_level': notification_level,
    }

  # Check the auto update frequency if they are enabled
  if notification_level_code == 4:
    # ScheduledInstallationTime is a default and unreliable value on
    # Windows 8 and Server 2012
    settings['installation_schedule'] = {
        'day': AUTO_UPDATE_SETTINGS['InstallationDay'][
            auto_client.get_item('Settings.ScheduledInstallationDay')],
        'time': AUTO_UPDATE_SETTINGS['InstallationTime'][
            auto_client.get_item('Settings.ScheduledInstallationTime')]
      }

  return settings



#!/usr/bin/env python
"""
Hiveary
https://hiveary.com

Licensed under Simplified BSD License (see LICENSE)
(C) Hiveary, LLC 2013 all rights reserved

Functions for finding and manipulating file paths.
"""

import os
import subprocess
import sys

# Windows specific imports
if subprocess.mswindows:
  from ctypes import c_int, wintypes, windll


CSIDL_COMMON_APPDATA = 35   # Used for finding the path to the config file in Windows


def get_common_appdata_path():
  """Returns the path to the common appdata folder in Windows. Typically this is
  either C:\ProgramData (Vista and higher) or
  C:\Documents and Settings\All Users\Application Data (XP).

  Returns:
    A string of the path to the common appdata folder.
  """

  _SHGetFolderPath = windll.shell32.SHGetFolderPathW
  _SHGetFolderPath.argtypes = [wintypes.HWND,
                               c_int,
                               wintypes.HANDLE,
                               wintypes.DWORD, wintypes.LPCWSTR]

  path_buf = wintypes.create_unicode_buffer(wintypes.MAX_PATH)
  _SHGetFolderPath(0, CSIDL_COMMON_APPDATA, 0, 0, path_buf)

  return path_buf.value


def get_program_path():
  """Returns the agent's directory, even if the agent is frozen.

  Returns:
    A string of the agent's directory.
  """

  if hasattr(sys, 'frozen'):
    resource_dir = os.path.dirname(
        unicode(sys.executable, sys.getfilesystemencoding()))
  else:
    resource_dir = os.path.dirname(
        unicode(__file__, sys.getfilesystemencoding()))

  return os.path.abspath(resource_dir)


def find_executable():
  """Determines the absolute path to the current process's executable and args.

  Returns:
    A tuple with the executable as the first and all args as a list for
    the second item.
  """

  if hasattr(sys, 'frozen'):
    # Figure out the right file to call. When frozen, the running executable will
    # be in a versioned subdirectory of the application.
    app_exe = os.path.basename(sys.executable)
    ver_dir = os.path.dirname(sys.executable)
    app_dir = os.path.dirname(ver_dir)

    # Make sure we aren't already at the root - in some instances there is
    # another directory to transision.
    if os.path.basename(app_dir) == 'appdata':
      app_dir = os.path.dirname(app_dir)

    executable = os.path.join(app_dir, app_exe)
    args = sys.argv[1:]
  else:
    executable = sys.executable
    args = sys.argv

    file_path = os.path.abspath(__file__)
    file_name = os.path.basename(file_path)

    # Check if the agent was invoked with the python command, instead
    # of just being run directly as an executable. If so, make sure that the
    # the absolute path is in the args.
    if file_name in args[0]:
      args[0] = file_path

  return (executable, args)



#!/usr/bin/env python

"""
Hiveary
https://hiveary.com

Licensed under Simplified BSD License (see LICENSE)
(C) Hiveary, LLC 2013 all rights reserved
"""

import platform
import sys

from hiveary import __version__ as version


current_platform = platform.system()

FROZEN_NAME = 'hiveary-agent'
AUTHOR = "Hiveary"
AUTHOR_EMAIL = "info@hiveary.com"
DESCRIPTION = "Hiveary Monitoring Agent"
LICENSE = "Simplified BSD"
URL = "http://hiveary.com"


# OS-specific setup
if 'bdist_esky' in sys.argv and current_platform == 'Windows':
  # Use esky/cxfreeze to build the agent and py2exe to build the service
  from esky.bdist_esky import Executable
  from glob import glob
  import os
  import py2exe  # noqa
  import setuptools
  import shutil

  modules = [
      'kombu.transport.pyamqp',
      'kombu.transport.base',
      'kombu.transport.amqplib',
  ]

  sys.path.append('C:\\Program Files (x86)\\Microsoft Visual Studio 9.0\\VC\\redist\\x86\\Microsoft.VC90.CRT')

  # Add in Visual Studio C++ compiler library
  data_files = [
      ('Microsoft.VC90.CRT', glob(r'C:\Program Files (x86)\Microsoft Visual Studio 9.0\VC\redist\x86\Microsoft.VC90.CRT\*.*')),
      r'hiveary\ca-bundle.pem',
      ('monitors', glob(r'monitors\*.py'))
  ]

  script = Executable('hiveary-agent', gui_only=False)

  options = {
      'bdist_esky': {
          'freezer_module': 'cxfreeze',
          'includes': modules,
      }
  }

  # Build the agent
  setuptools.setup(name=FROZEN_NAME,
                   version=version,
                   scripts=[script],
                   options=options,
                   data_files=data_files,
                   )

  sys.argv.remove('bdist_esky')
  sys.argv.append('py2exe')

  # used for the versioninfo resource
  class Target(object):
    def __init__(self, **kw):
      self.__dict__.update(kw)
      self.version = version
      self.company_name = 'Hiveary'
      self.name = "HivearyService"

  script = Target(
      description='Hiveary Agent Service Launcher',
      modules=["HivearyService"],
      cmdline_style='pywin32',
      )

  data_files = []

  # Build the service
  setuptools.setup(name='HivearyService',
                   version=version,
                   options={'py2exe': {}},
                   service=[script]
                   )

  # python27.dll will be available at the root once the esky zip is extracted,
  # so we can remove it now
  os.remove(r'dist\python27.dll')
  shutil.rmtree('build')

else:
  try:
    from setuptools import setup, find_packages
  except ImportError:
    from ez_setup import use_setuptools
    use_setuptools()
    from setuptools import setup, find_packages

  # Include all files from the package.
  install_requires = [
      'kombu>=2.5.7',
      'netifaces>=0.7',
      'oauth2>=1.5.211',
      'psutil>=1.1.0',
      'simplejson>=3.0.5',
      'Twisted>=12.2.0',
  ]

  data_files = [
      ('/etc/hiveary', ['hiveary.conf.example', 'README.md']),
      ('/etc/hiveary/init', ['initd/hiveary-agent']),
      ('/etc/hiveary/systemd', ['arch/hiveary-agent.service']),
      ('/usr/lib/hiveary', ['monitors/resources.py']),
  ]

  setup(name=FROZEN_NAME,
        version=version,
        author=AUTHOR,
        author_email=AUTHOR_EMAIL,
        description=DESCRIPTION,
        license=LICENSE,
        url=URL,
        include_package_data=True,
        data_files=data_files,
        install_requires=install_requires,
        packages=find_packages(),
        scripts=['hiveary-agent']
        )
